#include <iostream>
#include <unistd.h>
#include "Node.hpp"
#include "BarnesHutTreeThreaded.hpp"
#include "ContinuousPthreadSimulation.hpp"

namespace nbody {
	ContinuousPthreadSimulation::ContinuousPthreadSimulation() {
		this->iterations = 1;
	}

	ContinuousPthreadSimulation::~ContinuousPthreadSimulation() {
	}

	void ContinuousPthreadSimulation::initialize(string inputFile) {
		if (!this->readInputData(inputFile)) {
			cerr << "input file error" << endl;
			exit(-1);
		}
		int numThreads = sysconf(_SC_NPROCESSORS_ONLN);
		for (int i = 0; i < numThreads; i++) {
			pthread_t thread;

			this->threads.push_back(thread);
		}
		pthread_barrier_init(&this->barrier, NULL, this->threads.size());
		this->tree = new BarnesHutTreeThreaded(0);
		this->tree->simulation = this;
	}

	void ContinuousPthreadSimulation::cleanup() {
		pthread_barrier_destroy(&this->barrier);
		PthreadSimulation::cleanup();
	}

	void ContinuousPthreadSimulation::setIterations(int iterations) {
		this->iterations = iterations;
	}

	void* ContinuousPthreadSimulation::run(void* data) {
		NodesToProcess* toProcess = (NodesToProcess*) data;

		//build tree
		while (!BarnesHutTreeThreaded::hasNodeProcessingFinished(toProcess)) {
			Node* current = BarnesHutTreeThreaded::getNodeToProcess(toProcess);

			if (current != NULL) {
				BarnesHutTreeThreaded::updateNode(current);
				if (BarnesHutTree::splitNode(current)) {
					Node* child = current->next;

					while (child != NULL) {
						BarnesHutTreeThreaded::addNodeToProcess(child, toProcess);
						child = child->nextSibling;
					}
				}
			}
			BarnesHutTreeThreaded::checkNodeProcessingFinished(toProcess);
		}
		pthread_barrier_wait(toProcess->barrier);
		for (int i = 0; i < toProcess->iterations; i++) {
			vector<Body> extractBodies;

			//compute forces and move bodies
			while (!BarnesHutTreeThreaded::hasNodeProcessingFinished(toProcess)) {
				Node* current = BarnesHutTreeThreaded::getNodeToProcess(toProcess);

				if (current != NULL) {
					for (vector<Body>::iterator it = current->bodies.begin(); it != current->bodies.end(); it++) {
						if (!it->refinement) {
							toProcess->tree->accumulateForceOnto(*it);
						}
					}
				}
				BarnesHutTreeThreaded::checkNodeProcessingFinished(toProcess);
			}
			pthread_barrier_wait(toProcess->barrier);
			extractBodies.clear();
			for (Node* n = toProcess->tree->nodes->next; n != toProcess->tree->nodes; n = n->next) {
				if (n->lock.writeLock()) {
					n->extractLocalBodiesTo(extractBodies);
					n->lock.unLock();
				}
			}
			for (vector<Body>::iterator it = extractBodies.begin(); it != extractBodies.end(); it++) {
				integrate(*it);
			}
			if (pthread_barrier_wait(toProcess->barrier) == PTHREAD_BARRIER_SERIAL_THREAD) {
				toProcess->tree->simulation->clearBodies();
				toProcess->tree->clean();
				toProcess->tree->nodes->insertAfter(new Node(toProcess->tree));
				initBox(toProcess->tree->nodes->next->bb);
				toProcess->tree->nodes->next->afterSubtree = toProcess->tree->nodes->next->next;
				toProcess->toProcess.push_back(toProcess->tree->nodes->next);
			}
			pthread_barrier_wait(toProcess->barrier);
			pthread_mutex_lock(&toProcess->mutex);
			toProcess->tree->nodes->next->bodies.insert(toProcess->tree->nodes->next->bodies.end(), extractBodies.begin(), extractBodies.end());

			extendForBodies(toProcess->tree->nodes->next->bb, extractBodies);
			toProcess->tree->nodes->next->extendBBtoCube();
			pthread_mutex_unlock(&toProcess->mutex);
			pthread_barrier_wait(toProcess->barrier);
		}
		pthread_exit(NULL);
	}

	void ContinuousPthreadSimulation::run() {
		Box domain;
		BarnesHutTreeThreaded* bhtTree = (BarnesHutTreeThreaded*) this->tree;

		initBox(domain);
		bhtTree->init(this->bodies, domain);
		bhtTree->clearNodesToProcess();
		bhtTree->nodesToProcess.tree = bhtTree;
		bhtTree->nodesToProcess.barrier = &this->barrier;
		bhtTree->nodesToProcess.iterations = this->iterations;
		bhtTree->nodesToProcess.toProcess.push_back(bhtTree->nodes->next);
		for (int i = 0; i < this->getNumberOfProcesses(); i++) {
			pthread_create(&this->threads[i], NULL, ContinuousPthreadSimulation::run, &bhtTree->nodesToProcess);
		}
		for (int i = 0; i < this->getNumberOfProcesses(); i++) {
			void* retVal;

			pthread_join(this->threads[i], &retVal);
		}
		/*
		if (this->tree->isCorrect()) {
			cout << "correct" << endl;
		}
		*/
	}

}
